@@ -10,7 +10,7 @@
 #define KASTEN_BYTEARRAYSOURCECODESTREAMENCODERCONFIGEDITORFACTORY_HPP
 
 // lib
-#include <kasten/okteta/oktetakastengui_export.hpp>
+#include "oktetakastengui_export.hpp"
 // Kasten gui
 #include <Kasten/AbstractModelStreamEncoderConfigEditorFactory>
 
